from copy import copy
from enum import Enum
from pathlib import Path
from typing import Iterable, Mapping, Optional, Type, TypeVar, cast

import h5py
import numpy as np
from mpi4py import MPI

from opencosmo.header import OpenCosmoHeader
from opencosmo.index import ChunkedIndex
from opencosmo.mpi import get_comm_world

from .protocols import DataSchema
from .schemas import (
    ColumnSchema,
    DatasetSchema,
    EmptyColumnSchema,
    FileSchema,
    IdxLinkSchema,
    LightconeSchema,
    LinkSchema,
    SimCollectionSchema,
    SpatialIndexLevelSchema,
    SpatialIndexSchema,
    StartSizeLinkSchema,
    StructCollectionSchema,
    ZeroLengthError,
)

"""
When working with MPI, datasets are chunked across ranks. Here we combine the schemas
from several ranks into a single schema that can be allocated by rank 0. Each 
rank will then write it's own data to the specific section of the file 
it is responsible for.

As with schemas and writers, everything is very hierarcical here. A function
does some consistency checks, then calls a function that combines its children.
"""


class CombineState(Enum):
    VALID = 1
    ZERO_LENGTH = 2
    INVALID = 3


def write_parallel(file: Path, file_schema: FileSchema):
    comm = get_comm_world()
    if comm is None:
        raise ValueError("Got a null comm!")
    paths = set(comm.allgather(file))
    if len(paths) != 1:
        raise ValueError("Different ranks recieved a different path to output to!")

    try:
        file_schema.verify()
        results = comm.allgather(CombineState.VALID)
    except ValueError:
        results = comm.allgather(CombineState.INVALID)
    except ZeroLengthError:
        results = comm.allgather(CombineState.ZERO_LENGTH)
    if not all(results):
        raise ValueError("One or more ranks recieved invalid schemas!")

    has_data = [i for i, state in enumerate(results) if state == CombineState.VALID]
    group = comm.Get_group()
    new_group = group.Incl(has_data)
    new_comm = comm.Create(new_group)
    if new_comm == MPI.COMM_NULL:
        return cleanup_mpi(comm, new_comm, new_group)
    rank = new_comm.Get_rank()

    new_schema = combine_file_schemas(file_schema, new_comm)
    if rank == 0:
        with h5py.File(file, "w") as f:
            new_schema.allocate(f)

    writer = new_schema.into_writer(new_comm)

    try:
        with h5py.File(file, "a", driver="mpio", comm=new_comm) as f:
            writer.write(f)

    except ValueError:  # parallell hdf5 not available
        raise NotImplementedError(
            "MPI writes without paralell hdf5 are not yet supported"
        )
        nranks = new_comm.Get_size()
        rank = new_comm.Get_rank()
        for i in range(nranks):
            if i == rank:
                with h5py.File(file, "a") as f:
                    writer.write(f)
            new_comm.Barrier()
    cleanup_mpi(comm, new_comm, new_group)


def cleanup_mpi(comm_world: MPI.Comm, comm_write: MPI.Comm, group_write: MPI.Group):
    comm_world.Barrier()
    if comm_write != MPI.COMM_NULL:
        comm_write.Free()
    group_write.Free()


def get_all_child_names(schema: DataSchema | None, comm: MPI.Comm):
    if schema is None:
        child_names = set()
    elif hasattr(schema, "children") and isinstance(schema.children, dict):
        child_names = set(schema.children.keys())
    else:
        child_names = set()
    all_child_names: Iterable[str]
    all_child_names = child_names.union(*comm.allgather(child_names))
    all_child_names = list(all_child_names)
    all_child_names.sort()
    return all_child_names


def verify_structure(schemas: Mapping[str, DataSchema], comm: MPI.Comm):
    verify_names(schemas, comm)
    verify_types(schemas, comm)


def verify_names(schemas: Mapping[str, DataSchema], comm: MPI.Comm):
    names = set(schemas.keys())
    all_names = comm.allgather(names)
    if not all(ns == all_names[0] for ns in all_names[1:]):
        raise ValueError(
            "Tried to combine a collection of schemas with different names!"
        )


def verify_types(schemas: Mapping[str, DataSchema], comm: MPI.Comm):
    types = list(str(type(c)) for c in schemas.values())
    types.sort()
    all_types = comm.allgather(types)
    if not all(ts == all_types[0] for ts in all_types[1:]):
        raise ValueError(
            "Tried to combine a collection of schemas with different types!"
        )


def combine_file_schemas(schema: FileSchema, comm: MPI.Comm) -> FileSchema:
    if comm.Get_size() == 1:
        return schema

    all_child_names = get_all_child_names(schema, comm)
    new_schema = FileSchema()

    for child_name in all_child_names:
        child = schema.children.get(child_name)
        new_child = combine_file_child(child, comm)
        new_schema.add_child(new_child, child_name)

    return new_schema


S = TypeVar("S", DatasetSchema, SimCollectionSchema, StructCollectionSchema)


def combine_file_child(schema: S | None, comm: MPI.Comm) -> S:
    match schema:
        case DatasetSchema():
            return cast(S, combine_dataset_schemas(schema, comm))
        case SimCollectionSchema():
            return cast(S, combine_simcollection_schema(schema, comm))
        case StructCollectionSchema():
            return cast(S, combine_structcollection_schema(schema, comm))
        case LightconeSchema():
            return cast(S, combine_lightcone_schema(schema, comm))
        case _:
            raise ValueError(f"Invalid file child of type {type(schema)}")


def validate_headers(
    header: OpenCosmoHeader | None, comm: MPI.Comm, header_updates: dict = {}
):
    all_headers: Iterable[OpenCosmoHeader] = comm.allgather(header)
    all_headers = filter(lambda h: h is not None, all_headers)
    all_headers = list(map(lambda h: h.with_parameters(header_updates), all_headers))
    regions = set([h.file.region for h in all_headers])
    if len(regions) > 1:
        all_headers = [h.with_region(None) for h in all_headers]

    if any(h != all_headers[0] for h in all_headers[1:]):
        raise ValueError("Not all datasets have the same header!")
    return all_headers[0]


def combine_dataset_schemas(
    schema: DatasetSchema | None,
    comm: MPI.Comm,
    header_updates: dict = {},
) -> DatasetSchema:
    if schema is not None:
        header = validate_headers(schema.header, comm, header_updates)
        columns = schema.columns
    else:
        header = validate_headers(None, comm, header_updates)
        columns = {}

    new_schema = DatasetSchema(header=header)
    all_column_names = get_all_child_names(schema, comm)

    for colname in all_column_names:
        column = columns.get(colname)
        assert not isinstance(column, EmptyColumnSchema)
        new_column_schema = combine_column_schemas(column, comm)
        new_schema.columns[colname] = new_column_schema

    new_links = combine_links(schema.links if schema is not None else {}, comm)
    for name, link in new_links.items():
        new_schema.add_child(link, name)

    new_spatial_idx_schema = combine_spatial_index_schema(
        schema.spatial_index if schema is not None else None, comm
    )
    if new_spatial_idx_schema is not None:
        new_schema.add_child(new_spatial_idx_schema, "index")

    return new_schema


def combine_spatial_index_schema(
    schema: Optional[SpatialIndexSchema], comm: MPI.Comm = MPI.COMM_WORLD
):
    has_schema = schema is not None
    all_has_schema = comm.allgather(has_schema)

    if not any(all_has_schema):
        return None

    n_levels = max(schema.levels) if schema is not None else -1
    all_max_levels = set(comm.allgather(n_levels))
    if -1 in all_max_levels:
        all_max_levels.remove(-1)

    if len(set(all_max_levels)) != 1:
        raise ValueError("Schemas for all ranks must have the same number of levels!")

    max_level = all_max_levels.pop()
    level_schemas = schema.levels if schema is not None else {}
    new_schema = SpatialIndexSchema()
    for level in range(max_level + 1):
        level_schema = level_schemas.get(level)
        new_level_schema = combine_spatial_index_level_schemas(level_schema, comm)
        new_schema.add_child(new_level_schema, level)

    return new_schema


def combine_spatial_index_level_schemas(
    schema: SpatialIndexLevelSchema | None, comm: MPI.Comm
):
    if schema is None:
        start_len = 0
        size_len = 0
    else:
        start_len = len(schema.start)
        size_len = len(schema.size)

    all_start_lens = set(filter(lambda s: s is not None, comm.allgather(start_len)))
    all_size_lens = set(filter(lambda s: s is not None, comm.allgather(size_len)))
    if 0 in all_start_lens:
        all_start_lens.remove(0)
        all_size_lens.remove(0)

    if all_start_lens != all_size_lens or len(all_start_lens) != 1:
        raise ValueError("Invalid starts and sizes")

    level_len = all_start_lens.pop()

    if schema is None:
        source = np.zeros(level_len, dtype=np.int32)
        index = ChunkedIndex.from_size(len(source))
        start = ColumnSchema("start", index, source, {}, total_length=level_len)
        size = ColumnSchema("size", index, source, {}, total_length=level_len)
        new_schema = SpatialIndexLevelSchema(start, size)
    else:
        start = ColumnSchema(
            "start",
            schema.start.index,
            schema.start.source,
            {},
            total_length=level_len,
        )
        size = ColumnSchema(
            "size",
            schema.size.index,
            schema.size.source,
            {},
            total_length=level_len,
        )
        new_schema = SpatialIndexLevelSchema(start, size)

    return new_schema


def combine_links(
    links: dict[str, LinkSchema], comm: MPI.Comm
) -> Mapping[str, LinkSchema]:
    link_names = set(links.keys())

    all_link_names: Iterable[str]

    all_link_types: Iterable[Type]
    all_link_names = link_names.union(*comm.allgather(link_names))
    all_link_names = list(all_link_names)
    all_link_names.sort()
    new_links: dict[str, LinkSchema] = {}
    for link_name in all_link_names:
        link = links.get(link_name)
        all_link_types = comm.allgather(type(link))
        all_link_types = set(filter(lambda t: t is not type(None), all_link_types))
        if len(all_link_types) != 1:
            raise ValueError("Incompatible Links!")

        link_type = all_link_types.pop()

        if link_type is StartSizeLinkSchema:
            assert not isinstance(link, (IdxLinkSchema, EmptyColumnSchema))
            new_links[link_name] = combine_start_size_link_schema(link, comm, link_name)
        else:
            assert not isinstance(link, (StartSizeLinkSchema, EmptyColumnSchema))
            new_links[link_name] = combine_idx_link_schema(link, comm)

    return new_links


def combine_idx_link_schema(
    schema: IdxLinkSchema | None, comm: MPI.Comm
) -> IdxLinkSchema:
    column = schema.column if schema is not None else None
    assert not isinstance(column, EmptyColumnSchema)
    column_schema = combine_column_schemas(column, comm)
    new_schema = IdxLinkSchema(column_schema)
    return new_schema


def combine_start_size_link_schema(
    schema: StartSizeLinkSchema | None, comm: MPI.Comm, name: str
) -> StartSizeLinkSchema:
    start = schema.start if schema is not None else None
    size = schema.size if schema is not None else None
    assert not isinstance(start, EmptyColumnSchema) and not isinstance(
        size, EmptyColumnSchema
    )

    start_column_schema = combine_column_schemas(start, comm)
    size_column_schema = combine_column_schemas(size, comm)

    if schema is None:
        new_schema = StartSizeLinkSchema(name, start_column_schema, size_column_schema)
    else:
        new_schema = copy(schema)
        new_schema.start = start_column_schema
        new_schema.size = size_column_schema
    return new_schema


def combine_lightcone_schema(schema: LightconeSchema | None, comm: MPI.Comm):
    all_child_names = get_all_child_names(schema, comm)
    new_schema = LightconeSchema()
    if schema is None:
        children = {}
    else:
        children = schema.children

    for child_name in all_child_names:
        child = children.get(child_name)
        z_range = get_z_range(child, comm)

        new_dataset_schema = combine_dataset_schemas(
            children.get(child_name), comm, {"lightcone/z_range": z_range}
        )
        new_schema.add_child(new_dataset_schema, child_name)
    return new_schema


def get_z_range(ds: DatasetSchema | None, comm: MPI.Comm):
    if ds is not None and ds.header is not None:
        z_ranges = comm.allgather(ds.header.lightcone["z_range"])
    else:
        z_ranges = comm.allgather(None)
    z_ranges = list(filter(lambda dz: dz is not None, z_ranges))
    dzs: Iterable[float] = map(lambda dz: dz[1] - dz[0], z_ranges)
    dzs = list(dzs)
    max_idx = np.argmax(dzs)
    return list(z_ranges)[max_idx]


def combine_simcollection_schema(
    schema: SimCollectionSchema, comm: MPI.Comm
) -> SimCollectionSchema:
    verify_structure(schema.children, comm)

    child_names = schema.children.keys()

    new_schema = SimCollectionSchema()
    new_child: DatasetSchema | StructCollectionSchema

    for child_name in child_names:
        child_name_ = comm.bcast(child_name)
        child = schema.children[child_name_]
        match child:
            case StructCollectionSchema():
                new_child = combine_structcollection_schema(child, comm)
            case DatasetSchema():
                new_child = combine_dataset_schemas(child, comm)
        new_schema.add_child(new_child, child_name)
    return new_schema


def combine_structcollection_schema(
    schema: StructCollectionSchema, comm: MPI.Comm
) -> StructCollectionSchema:
    child_names: Iterable[str] = set(schema.children.keys())
    all_child_names = comm.allgather(child_names)
    if not all(cns == all_child_names[0] for cns in all_child_names[1:]):
        raise ValueError(
            "Tried to combine ismulation collections with different children!"
        )

    child_types = set(str(type(c)) for c in schema.children.values())
    all_child_types = comm.allgather(child_types)
    if not all(cts == all_child_types[0] for cts in all_child_types[1:]):
        raise ValueError(
            "Tried to combine ismulation collections with different children!"
        )

    new_schema = StructCollectionSchema()
    child_names = list(child_names)
    child_names.sort()
    new_child: DatasetSchema | StructCollectionSchema

    for i, name in enumerate(child_names):
        cn = comm.bcast(name)
        child = schema.children[cn]
        if isinstance(child, DatasetSchema):
            new_child = combine_dataset_schemas(child, comm)
        elif isinstance(child, StructCollectionSchema):
            new_child = combine_structcollection_schema(child, comm)
        else:
            raise ValueError(
                "Found a child of a structure collection that was not a Dataset!"
            )
        new_schema.add_child(new_child, cn)

    return new_schema


def verify_column_schemas(schema: ColumnSchema | None, comm: MPI.Comm):
    if schema is None:
        data = comm.allgather(None)
    else:
        data = comm.allgather(
            (schema.source.shape, schema.name, dict(schema.attrs), schema.source.dtype)
        )

    data = list(filter(lambda elem: elem is not None, data))
    if any(d[1:] != data[0][1:] for d in data[1:]):
        raise ValueError("Tried to write incompatible columns to the same output!")
    return data[0]


def combine_column_schemas(
    schema: ColumnSchema | None, comm: MPI.Comm
) -> ColumnSchema | EmptyColumnSchema:
    rank = comm.Get_rank()
    if schema is None:
        length = 0
    else:
        length = len(schema.index)

    shape, name, attrs, dtype = verify_column_schemas(schema, comm)

    lengths = comm.allgather(length)
    total_length = np.sum(lengths)
    rank_offsets = np.insert(np.cumsum(lengths), 0, 0)[:-1]
    rank_offset = rank_offsets[rank]

    new_schema: ColumnSchema | EmptyColumnSchema
    if schema is None:
        new_schema = EmptyColumnSchema(name, attrs, dtype, (total_length,) + shape[1:])
    else:
        new_schema = ColumnSchema(
            schema.name,
            schema.index,
            schema.source,
            schema.attrs,
            total_length=total_length,
        )
        if length != 0:
            new_schema.set_offset(rank_offset)
    return new_schema



from __future__ import annotations

from typing import TYPE_CHECKING

import h5py
import numpy as np

from opencosmo.index import ChunkedIndex, DataIndex, SimpleIndex
from opencosmo.io import schemas as ios

if TYPE_CHECKING:
    pass


class LinkedDatasetHandler:
    """
    Links are currently only supported out-of-memory.
    """

    def __init__(
        self,
        link: h5py.Group | tuple[h5py.Group, h5py.Group],
    ):
        self.link = link

    def has_linked_data(self, index: DataIndex) -> np.ndarray:
        """
        Check which rows in this index actually have data
        """
        if isinstance(self.link, tuple):
            sizes = index.get_data(self.link[1])
            return sizes > 0

        else:
            rows = index.get_data(self.link)
            return rows != -1

    def make_index(self, index: DataIndex) -> DataIndex:
        if isinstance(self.link, tuple):
            start = index.get_data(self.link[0])
            size = index.get_data(self.link[1])
            valid_rows = size > 0
            start = start[valid_rows]
            size = size[valid_rows]
            if not start.size:
                return SimpleIndex(np.array([], dtype=int))
            else:
                return ChunkedIndex(start, size)
        else:
            indices_into_data = index.get_data(self.link)
            indices_into_data = indices_into_data[indices_into_data >= 0]

            return SimpleIndex(indices_into_data)

    def make_schema(self, name: str, index: DataIndex) -> ios.LinkSchema:
        if isinstance(self.link, h5py.Dataset):
            return ios.IdxLinkSchema.from_h5py_dataset(name, index, self.link)
        else:
            return ios.StartSizeLinkSchema.from_h5py_dataset(name, index, *self.link)



from typing import Callable, ClassVar

from pydantic import BaseModel, Field, computed_field

from opencosmo.cosmology import make_cosmology


class CosmologyParameters(BaseModel, frozen=True):
    """
    Responsible for validating cosmology parameters and handling differences in
    naming conventions between OpenCosmo and astropy.cosmology. Generally should
    not be used by the user directly
    """

    ACCESS_PATH: ClassVar[str] = "cosmology"
    ACCESS_TRANSFORMATION: ClassVar[Callable] = make_cosmology

    h: float = Field(ge=0.0, description="Reduced Hubble constant")

    @computed_field  # type: ignore
    @property
    def H0(self) -> float:
        """
        Hubble constant in km/s/Mpc
        """
        return self.h * 100

    Om0: float = Field(ge=0.0, description="Total matter density", alias="omega_m")
    Ob0: float = Field(ge=0.0, description="Baryon density", alias="omega_b")
    Ode0: float = Field(ge=0.0, description="Dark energy density", alias="omega_l")
    Neff: float = Field(
        gt=0.0, description="Effective number of neutrinos", alias="n_eff_massless"
    )
    n_eff_massive: float = Field(
        0, ge=0.0, description="Effective number of massive neutrinos"
    )
    sigma_8: float = Field(ge=0.0, description="RMS mass fluctuation at 8 Mpc/h")
    w0: float = Field(description="Dark energy equation of state", alias="w_0")
    wa: float = Field(
        description="Dark energy equation of state evolution", alias="w_a"
    )



from typing import Iterable, Optional, Protocol, Self

import h5py

from opencosmo.dataset import Dataset
from opencosmo.header import OpenCosmoHeader
from opencosmo.index import DataIndex


class LinkHandler(Protocol):
    """
    A LinkHandler is responsible for handling linked datasets. Links are found
    in property files, and contain indexes into another dataset. For example, a
    halo properties file will contain links to a halo particles file. Each halo
    in the properties file will have a corresponding range of indexes that contain
    the associated particles in the particles file.

    The link handler is responsible for reading data and instatiating datasets
    that contain the linked data for the given object. There will be one link
    handler for each linked dataset in the properties file. This potentially
    means there will be multiple pointers to a single particle file, for example.
    """

    def __init__(
        self,
        file: h5py.File | h5py.Group,
        link: h5py.Group | tuple[h5py.Group, h5py.Group],
        header: OpenCosmoHeader,
        builder: Optional["DatasetBuilder"] = None,
        **kwargs,
    ):
        """
        Initialize the LinkHandler with the file, link, header, and optional builder.
        The builder is used to build the dataset from the file.
        """
        pass

    def get_data(self, index: DataIndex) -> Dataset:
        """
        Given a index or a set of indices, return the data from the linked dataset
        that corresponds to the halo/galaxy at that index in the properties file.
        Sometimes the linked dataset will not have data for that object, in which
        a zero-length dataset will be returned.
        """
        pass

    def get_all_data(self) -> Dataset:
        """
        Return all the data from the linked dataset.
        """
        pass

    def prep_write(
        self,
        data_group: h5py.Group,
        link_group: h5py.Group,
        name: str,
        index: DataIndex,
    ) -> None:
        """
        Write the linked data for the given indices to data_group.
        This function will then update the links to be consistent with the newly
        written data, and write the updated links to link_group.
        """
        pass

    def select(self, columns: str | Iterable[str]) -> Self:
        """
        Return a new LinkHandler that only contains the data for the given indices.
        """
        pass

    def with_units(self, convention: str) -> Self:
        """
        Return a new LinkHandler that uses the given unit convention.
        """
        pass


class DatasetBuilder(Protocol):
    """
    A DatasetBuilder is responsible for building a dataset from a file. It
    contains the logic for selecting columns and applying transformations to
    the data.
    """

    def with_units(self, convention: str) -> Self:
        pass

    def select(self, selected: str | Iterable[str]) -> Self:
        pass

    def build(
        self,
        file: h5py.File | h5py.Group,
        header: OpenCosmoHeader,
        index: Optional[DataIndex] = None,
    ) -> Dataset:
        pass



import h5py

from opencosmo.file import broadcast_read
from opencosmo.transformations import protocols as t


@broadcast_read
def generate_transformations(
    input: h5py.Group,
    generators: list[t.TransformationGenerator],
    existing: t.TransformationDict = {},
) -> t.TransformationDict:
    """
    Generate transformations based on the input dataset and a list of generators.
    Generated transformations will always be run before other transformations.

    The logic is that generators rely on data that will not be accessible after
    the data is moved into memory, and so in some sense "precede" transformations
    that only operate on the in-memory representation.
    """
    for dataset in input.values():
        for gen in generators:
            generated_transformations = gen(dataset)
            if generated_transformations is not None:
                for (
                    transformation_type,
                    transformations,
                ) in generated_transformations.items():
                    existing_transformation = existing.get(transformation_type, [])
                    existing[transformation_type] = list(transformations) + list(
                        existing_transformation
                    )
    return existing



from enum import Enum
from typing import Optional, Protocol, Sequence, runtime_checkable

import h5py
from astropy.table import Table  # type: ignore
from numpy.typing import ArrayLike


class TransformationType(Enum):
    TABLE = "table"
    COLUMN = "column"
    ALL_COLUMNS = "all_columns"
    FILTER = "filter"


class TransformationGenerator(Protocol):
    """
    A transformation generator is a callable that returns a transformation
    or set of transformations based on information stored in the attributes of a given
    dataset. Examples include units stored as attributes
    """

    def __call__(self, column: h5py.Dataset) -> "TransformationDict": ...


class TableTransformation(Protocol):
    """
    A transformation that can be applied to a table, producing a new table.

    The new table will replace the original table.
    """

    def __call__(self, input: Table) -> Table: ...


@runtime_checkable
class AllColumnTransformation(Protocol):
    """
    A transformation that is applied to all columns in a table.
    """

    def __call__(self, input: ArrayLike) -> ArrayLike: ...


class ColumnTransformation(Protocol):
    """
    A transformation that is applied to a single column, producing
    an updated version of that version of that column.

    An "all_columns" transformation is just a regular column transformation
    except that it will be applied to all columns in the table. In this case,
    column_name should return None.
    """

    def __init__(self, column_name: str, *args, **kwargs): ...

    @property
    def column_name(self) -> Optional[str]: ...

    def __call__(self, input: ArrayLike) -> ArrayLike: ...


Transformation = TableTransformation | ColumnTransformation | AllColumnTransformation
TransformationDict = dict[TransformationType, Sequence[Transformation]]



from typing import Callable, Optional

import h5py
import numpy as np

from opencosmo.header import OpenCosmoHeader
from opencosmo.index import DataIndex, SimpleIndex
from opencosmo.io import protocols as iop

try:
    from mpi4py import MPI
except ImportError:
    MPI = None  # type: ignore

"""
Writers work in tandem with schemas to create new files. All schemas must have
an into_writer method, which returns a writer that can be used to put
data into the new file.

Schemas are responsible for validating and building the file structure 
as well as allocating space.  As a result, writers ASSUME the correct structure exists, 
and that all the datasets have the correct size, datatype, etc.
"""


def write_index(
    input_ds: h5py.Dataset | np.ndarray | None,
    output_ds: h5py.Dataset,
    index: DataIndex,
    offset: int = 0,
    updater: Optional[Callable[[np.ndarray], np.ndarray]] = None,
):
    """
    Helper function to take elements from one h5py.Dataset using an index
    and put it in a different one.
    """
    data = np.array([])
    if len(index) > 0 and input_ds is not None:
        data = index.get_data(input_ds)
        if updater is not None:
            data = updater(data)

        data = data.astype(input_ds.dtype)

    if input_ds is not None:
        output_ds[offset : offset + len(data)] = data


class FileWriter:
    """
    Root writer for a file. Pretty much just calls the child writers.
    """

    def __init__(self, children: dict[str, iop.DataWriter]):
        self.children = children

    def write(self, file: h5py.File):
        if len(self.children) == 1:
            ds = next(iter(self.children.values()))
            return ds.write(file)
        for name, dataset in self.children.items():
            dataset.write(file[name])


class CollectionWriter:
    """
    Writes collections to a file or grous. Also pretty much just calls
    the child writers. May or may not recieve a header to write, depending
    on they type of collection.
    """

    def __init__(
        self,
        children: dict[str, iop.DataWriter],
        header: Optional[OpenCosmoHeader] = None,
    ):
        self.children = children
        self.header = header

    def write(self, file: h5py.File | h5py.Group):
        if len(self.children) == 1:
            next(iter(self.children.values())).write(file)
            return

        child_names = list(self.children.keys())
        child_names.sort()
        for name in child_names:
            self.children[name].write(file[name])


class DatasetWriter:
    """
    Writes datasets to a file or group. Datasets must have at least one column.
    If the datset is being written alone or as part of SimulationCollection, it will
    be responsible for writing a header.

    It may or may not have a spatial index. It also may or may not have links
    to other datasets.
    """

    def __init__(
        self,
        columns: dict[str, "ColumnWriter"],
        links: dict[str, "LinkWriter"] = {},
        spatial_index: Optional["SpatialIndexWriter"] = None,
    ):
        self.columns = columns
        self.links = links
        self.spatial_index = spatial_index

    def write(self, group: h5py.Group):
        data_group = group["data"]

        names = list(self.columns.keys())
        names.sort()
        for colname in names:
            self.columns[colname].write(data_group)
        if self.links:
            link_group = group["data_linked"]
            link_names = list(self.links.keys())
            link_names.sort()

            for name in link_names:
                self.links[name].write(link_group)

        if self.spatial_index is not None:
            index_group = group["index"]
            self.spatial_index.write(index_group)


class EmptyColumnWriter:
    def __init__(self, name: str):
        self.name = name

    def write(
        self,
        group: h5py.Group,
        updater: Optional[Callable[[np.ndarray], np.ndarray]] = None,
    ):
        ds = group[self.name]

        write_index(None, ds, SimpleIndex.empty(), updater=updater)


class ColumnWriter:
    """
    Writes a single column in a dataset. This is the only writer that actually moves
    real data around.
    """

    def __init__(
        self,
        name: str,
        index: DataIndex,
        source: h5py.Dataset,
        offset: int = 0,
    ):
        self.name = name
        self.source = source
        self.index = index
        self.offset = offset

    def write(
        self,
        group: h5py.Group,
        updater: Optional[Callable[[np.ndarray], np.ndarray]] = None,
    ):
        ds = group[self.name]

        write_index(self.source, ds, self.index, self.offset, updater)


class SpatialIndexWriter:
    """
    Writer for spatial indices. Mostly responsible for calling its children
    """

    def __init__(self, levels: dict[int, "SpatialIndexLevelWriter"]):
        self.levels = levels

    def write(self, group: h5py.Group):
        for (
            level_num,
            writer,
        ) in self.levels.items():
            level_group = group[f"level_{level_num}"]
            writer.write(level_group)


class SpatialIndexLevelWriter:
    """
    Writer for writing a single level of the spatial index. If this operation is being
    performed in an MPI context, the spatial indices must be summed.
    """

    def __init__(
        self, start: ColumnWriter, size: ColumnWriter, comm: Optional["MPI.Comm"] = None
    ):
        self.start = start
        self.size = size
        self.updater = lambda data: sum_updater(data, comm)

    def write(self, group: h5py.Group):
        self.size.write(group, updater=self.updater)
        self.start.write(group, updater=self.updater)


def sum_updater(data: np.ndarray, comm: Optional["MPI.Comm"] = None):
    if comm is not None and comm.Get_size():
        recvbuf = np.zeros_like(data)
        comm.Allreduce(data, recvbuf, MPI.SUM)
        return recvbuf
    return data


class IdxLinkWriter:
    """
    Writer for links between datasets, where each row in one dataset corresponds
    to a single row in the other. When the dataset is filtered, this link must be
    updated.
    """

    def __init__(self, col_writer: ColumnWriter, comm: Optional["MPI.Comm"] = None):
        self.writer = col_writer
        self.updater = make_idx_link_updater(self.writer, comm)

    def write(self, group: h5py.Group):
        self.writer.write(group, self.updater)


def idx_link_updater(input: np.ndarray, offset: int = 0) -> np.ndarray:
    output = np.full(len(input), -1)
    good = input >= 0
    output[good] = np.arange(sum(good)) + offset
    return output


def make_idx_link_updater(
    input: ColumnWriter, comm: Optional["MPI.Comm"]
) -> Callable[[np.ndarray], np.ndarray]:
    """
    Helper function to update data from a 1-to-1 index
    link.
    """
    arr = input.index.get_data(input.source)

    has_data = arr > 0
    offset = 0
    n_good = sum(has_data)
    if comm is not None:
        all_sizes = comm.allgather(n_good)
        offsets = np.insert(np.cumsum(all_sizes), 0, 0)
        offset = offsets[comm.Get_rank()]
    return lambda arr_: idx_link_updater(arr_, offset)


class StartSizeLinkWriter:
    """
    Writer for links between datasets where each row in one datest
    corresponds to several rows in the other.
    """

    def __init__(
        self, start: ColumnWriter, size: ColumnWriter, comm: Optional["MPI.Comm"] = None
    ):
        self.start = start
        self.sizes = size
        self.updater = make_start_link_updater(size, comm)

    def write(self, group: h5py.Group):
        self.sizes.write(group)
        new_sizes = self.sizes.index.get_data(self.sizes.source)
        self.start.write(group, lambda _: self.updater(new_sizes))


def start_link_updater(sizes: np.ndarray, offset: int = 0) -> np.ndarray:
    cumulative_sizes = np.cumsum(sizes)

    new_starts = np.insert(cumulative_sizes, 0, 0)
    new_starts = new_starts[:-1] + offset
    return new_starts


def make_start_link_updater(
    size_writer: ColumnWriter, comm: Optional["MPI.Comm"]
) -> Callable[[np.ndarray], np.ndarray]:
    """
    Helper function to update the starts of a start-size
    link.
    """
    sizes = size_writer.index.get_data(size_writer.source)
    total_size = np.sum(sizes)
    if comm is not None:
        offsets = np.cumsum(comm.allgather(total_size), dtype=np.uint64)
        offsets = np.insert(offsets, 0, 0)
        offset = offsets[comm.Get_rank()]
    else:
        offset = 0

    return lambda arr_: start_link_updater(arr_, offset)


LinkWriter = IdxLinkWriter | StartSizeLinkWriter



from collections import defaultdict
from pathlib import Path
from typing import Iterable

import h5py
import numpy as np
from deprecated import deprecated

from opencosmo import dataset as d
from opencosmo import io
from opencosmo.collection.structure import structure as sc
from opencosmo.header import OpenCosmoHeader
from opencosmo.index import DataIndex

from .handler import LinkedDatasetHandler

LINK_ALIASES = {  # Left: Name in file, right: Name in collection
    "sodbighaloparticles_star_particles": "star_particles",
    "sodbighaloparticles_dm_particles": "dm_particles",
    "sodbighaloparticles_gravity_particles": "gravity_particles",
    "sodbighaloparticles_agn_particles": "agn_particles",
    "sodbighaloparticles_gas_particles": "gas_particles",
    "sod_profile": "halo_profiles",
    "galaxyproperties": "galaxy_properties",
    "galaxyparticles_star_particles": "star_particles",
}

ALLOWED_LINKS = {  # h5py.Files that can serve as a link holder and
    "halo_properties": ["halo_particles", "halo_profiles", "galaxy_properties"],
    "galaxy_properties": ["galaxy_particles"],
}


@deprecated(
    version="0.8",
    reason="oc.open_linked_files is deprecated and will be removed in version 1.0. "
    "Please use oc.open instead",
)
def open_linked_files(*files: Path, **load_kwargs: bool):
    """
    **WARNING: THIS METHOD IS DEPCREATED AND WILL BE REMOVED IN A FUTURE
    VERSION. PLEASE USE** :py:meth:`opencosmo.open`

    Open a collection of files that are linked together, such as a
    properties file and a particle file.

    """
    if len(files) == 1 and isinstance(files[0], list):
        return open_linked_files(*files[0])

    return io.io.open(*files, **load_kwargs)


def validate_linked_groups(groups: dict[str, h5py.Group]):
    if "halo_properties" in groups:
        if "data_linked" not in groups["halo_properties"].keys():
            raise ValueError(
                "File appears to be a structure collection, but does not have links!"
            )
    elif "galaxy_properties" in groups:
        if "data_linked" not in groups["galaxy_properties"].keys():
            raise ValueError(
                "File appears to be a structure collection, but does not have links!"
            )
    if len(groups) == 1:
        raise ValueError("Structure collections must have more than one dataset")


def get_linked_datasets(
    linked_files_by_type: dict[str, h5py.File | h5py.Group],
    header: OpenCosmoHeader,
):
    targets = {}
    for dtype, pointer in linked_files_by_type.items():
        if "data" not in pointer.keys():
            targets.update(
                {
                    k: io.io.OpenTarget(pointer[k], header)
                    for k in pointer.keys()
                    if k != "header"
                }
            )
        else:
            targets.update({dtype: io.io.OpenTarget(pointer, header)})
    datasets = {
        dtype: io.io.open_single_dataset(target) for dtype, target in targets.items()
    }
    return datasets


def make_index_with_linked_data(
    index: DataIndex, links: dict[str, LinkedDatasetHandler]
):
    mask = np.ones(len(index), dtype=bool)
    for link in links.values():
        mask &= link.has_linked_data(index)

    return index.mask(mask)


def build_structure_collection(targets: list[io.io.OpenTarget], ignore_empty: bool):
    link_sources = defaultdict(list)
    link_targets: dict[str, dict[str, d.Dataset | sc.StructureCollection]] = (
        defaultdict(dict)
    )
    for target in targets:
        if target.data_type == "halo_properties":
            link_sources["halo_properties"].append(target)
        elif target.data_type == "galaxy_properties":
            link_sources["galaxy_properties"].append(target)
        elif target.data_type.startswith("halo"):
            dataset = io.io.open_single_dataset(target)
            name = target.group.name.split("/")[-1]
            if not name:
                name = target.data_type
            elif name.startswith("halo_properties"):
                name = name[16:]
            link_targets["halo_targets"][name] = dataset
        elif target.data_type.startswith("galaxy"):
            dataset = io.io.open_single_dataset(target)
            name = target.group.name.split("/")[-1]
            if not name:
                name = target.data_type
            elif name.startswith("galaxy_properties"):
                name = name[18:]
            link_targets["galaxy_targets"][name] = dataset
        else:
            raise ValueError(
                f"Unknown data type for structure collection {target.data_type}"
            )

    if len(link_sources["galaxy_properties"]) == 1 and link_targets["galaxy_targets"]:
        handlers = get_link_handlers(
            link_sources["galaxy_properties"][0].group,
            list(link_targets["galaxy_targets"].keys()),
            link_sources["galaxy_properties"][0].header,
        )

        source_dataset = io.io.open_single_dataset(link_sources["galaxy_properties"][0])
        if ignore_empty:
            new_index = make_index_with_linked_data(source_dataset.index, handlers)
            source_dataset = source_dataset.with_index(new_index)
        collection = sc.StructureCollection(
            source_dataset,
            source_dataset.header,
            link_targets["galaxy_targets"],
            handlers,
        )
        if len(link_sources["halo_properties"]) != 0:
            link_targets["halo_targets"]["galaxy_properties"] = collection
        else:
            return collection

    if len(link_sources["halo_properties"]) == 1 and link_targets["halo_targets"]:
        handlers = get_link_handlers(
            link_sources["halo_properties"][0].group,
            list(link_targets["halo_targets"].keys()),
            link_sources["halo_properties"][0].header,
        )
        source_dataset = io.io.open_single_dataset(link_sources["halo_properties"][0])

        if ignore_empty:
            new_index = make_index_with_linked_data(source_dataset.index, handlers)
            source_dataset = source_dataset.with_index(new_index)

        return sc.StructureCollection(
            source_dataset,
            source_dataset.header,
            link_targets["halo_targets"],
            handlers,
        )


def get_link_handlers(
    link_file: h5py.File | h5py.Group,
    linked_files: Iterable[str],
    header: OpenCosmoHeader,
) -> dict[str, LinkedDatasetHandler]:
    if "data_linked" not in link_file.keys():
        raise KeyError("No linked datasets found in the file.")
    links = link_file["data_linked"]

    linked_files = list(linked_files)
    unique_dtypes = {key.rsplit("_", 1)[0] for key in links.keys()}
    output_links = {}
    for dtype in unique_dtypes:
        if dtype not in linked_files and LINK_ALIASES.get(dtype) not in linked_files:
            continue  # Skip if the linked file is not provided

        key = LINK_ALIASES.get(dtype, dtype)
        try:
            start = links[f"{dtype}_start"]
            size = links[f"{dtype}_size"]

            output_links[key] = LinkedDatasetHandler(
                (start, size),
            )
        except KeyError:
            index = links[f"{dtype}_idx"]
            output_links[key] = LinkedDatasetHandler(index)
    return output_links



from __future__ import annotations

from typing import Iterable, Optional

import h5py
import numpy as np
from astropy.table import QTable  # type: ignore

from opencosmo.dataset.builders import TableBuilder
from opencosmo.header import OpenCosmoHeader
from opencosmo.index import DataIndex
from opencosmo.io.schemas import DatasetSchema
from opencosmo.mpi import get_comm_world
from opencosmo.transformations.units import get_raw_units


class DatasetHandler:
    """
    Handler for opencosmo.Dataset
    """

    def __init__(
        self,
        file: h5py.File,
        group_name: Optional[str] = None,
    ):
        self.__group_name = group_name
        self.__file = file
        if group_name is None:
            self.__group = file["data"]
        else:
            self.__group = file[f"{group_name}/data"]

    def __len__(self) -> int:
        first_column_name = next(iter(self.__group.keys()))
        return self.__group[first_column_name].shape[0]

    def __enter__(self):
        return self

    def __exit__(self, *exec_details):
        self.__group = None
        return self.__file.close()

    def collect(self, columns: Iterable[str], index: DataIndex) -> DatasetHandler:
        if (comm := get_comm_world()) is not None:
            indices = comm.allgather(index)
            new_index = indices[0].concatenate(*indices[1:])
        else:
            new_index = index
        file: h5py.File = h5py.File.in_memory()
        group = file.require_group("data")
        for colname in columns:
            dataset = self.__group[colname]
            data = new_index.get_data(dataset)
            group.create_dataset(colname, data=data)
            for name, value in dataset.attrs.items():
                group[colname].attrs[name] = value

        return DatasetHandler(file, group_name=self.__group_name)

    def get_raw_units(self, columns: Iterable[str]):
        return {col: get_raw_units(self.__group[col]) for col in columns}

    def prep_write(
        self,
        index: DataIndex,
        columns: Iterable[str],
        header: Optional[OpenCosmoHeader] = None,
    ) -> DatasetSchema:
        return DatasetSchema.make_schema(self.__group, columns, index, header)

    def get_data(self, builder: TableBuilder, index: DataIndex) -> QTable:
        """ """
        if self.__group is None:
            raise ValueError("This file has already been closed")
        data = {}
        for colname in builder.columns:
            data[colname] = index.get_data(self.__group[colname])

        return builder.build(data)

    def take_range(self, start: int, end: int, indices: np.ndarray) -> np.ndarray:
        if start < 0 or end > len(indices):
            raise ValueError("Indices out of range")
        return indices[start:end]



from typing import Iterable

import astropy.units as u  # type: ignore
import numpy as np
from numpy.typing import NDArray

from opencosmo.index import DataIndex


class InMemoryColumnHandler:
    def __init__(
        self,
        columns: dict[str, NDArray | u.Quantity],
        index: DataIndex,
    ):
        self.__index = index
        self.__columns = columns

    @classmethod
    def empty(cls, index: DataIndex):
        return InMemoryColumnHandler({}, index)

    def with_columns(self, columns: Iterable[str]):
        new_columns = {
            key: self.__columns[key] for key in columns if key in self.__columns
        }
        return InMemoryColumnHandler(new_columns, self.__index)

    def keys(self):
        return self.__columns.keys()

    def with_new_column(self, name: str, column: np.ndarray | u.Quantity):
        if len(column) != len(self.__index):
            raise ValueError("Tried to add an in-memory column with the wrong length!")
        new_columns = {**self.__columns, name: column}
        return InMemoryColumnHandler(new_columns, self.__index)

    def project(self, index: DataIndex):
        if len(self.__columns) == 0:
            return InMemoryColumnHandler.empty(index)
        index_into_columns = self.__index.projection(index)

        new_columns = {
            name: index_into_columns.get_data(col)
            for name, col in self.__columns.items()
        }
        return InMemoryColumnHandler(new_columns, index)

    def columns(self):
        yield from self.__columns.items()



from functools import reduce
from typing import TYPE_CHECKING, Iterable, Optional

import numpy as np
from astropy import table, units  # type: ignore
from astropy.cosmology import Cosmology  # type: ignore
from numpy.typing import NDArray

import opencosmo.transformations.units as u
from opencosmo.dataset.builders import TableBuilder, get_table_builder
from opencosmo.dataset.column import DerivedColumn
from opencosmo.dataset.im import InMemoryColumnHandler
from opencosmo.header import OpenCosmoHeader
from opencosmo.index import ChunkedIndex, DataIndex, SimpleIndex
from opencosmo.io import schemas as ios
from opencosmo.spatial.protocols import Region

if TYPE_CHECKING:
    from opencosmo.dataset.handler import DatasetHandler


class DatasetState:
    """
    Holds mutable state required by the dataset. Cleans up the dataset to mostly focus
    on very high-level operations. Not a user facing class.
    """

    def __init__(
        self,
        base_unit_transformations: dict,
        builder: TableBuilder,
        index: DataIndex,
        convention: u.UnitConvention,
        region: Region,
        header: OpenCosmoHeader,
        im_handler: InMemoryColumnHandler,
        sort_by: Optional[tuple[str, bool]] = None,
        hidden: set[str] = set(),
        derived: dict[str, DerivedColumn] = {},
    ):
        self.__base_unit_transformations = base_unit_transformations
        self.__builder = builder
        self.__im_handler = im_handler
        self.__convention = convention
        self.__derived: dict[str, DerivedColumn] = derived
        self.__header = header
        self.__hidden = hidden
        self.__index = index
        self.__sort_by = sort_by
        self.__region = region

    @property
    def index(self):
        return self.__index

    @property
    def builder(self):
        return self.__builder

    @property
    def convention(self):
        return self.__convention

    @property
    def region(self):
        return self.__region

    @property
    def header(self):
        return self.__header

    @property
    def columns(self) -> list[str]:
        columns = (
            set(self.__builder.columns)
            | set(self.__derived.keys())
            | set(self.__im_handler.keys())
        )
        return list(columns - self.__hidden)

    def get_data(
        self,
        handler: "DatasetHandler",
        ignore_sort: bool = False,
        attach_index: bool = False,
    ) -> table.QTable:
        """
        Get the data for a given handler.
        """
        data = handler.get_data(builder=self.__builder, index=self.__index)
        data = self.__get_im_columns(data)
        data = self.__build_derived_columns(data)
        data_columns = set(data.columns)
        index_array = self.__index.into_array()

        if not ignore_sort and self.__sort_by is not None:
            order = data.argsort(self.__sort_by[0], reverse=self.__sort_by[1])
            data = data[order]
            index_array = index_array[order]
        if (
            self.__hidden
            and not self.__hidden.intersection(data_columns) == data_columns
        ):
            data.remove_columns(self.__hidden)
        if attach_index:
            data["raw_index"] = index_array
        return data

    def with_index(self, index: DataIndex):
        """
        Return the same dataset state with a new index
        """

        new_cache = self.__im_handler.project(index)

        return DatasetState(
            self.__base_unit_transformations,
            self.__builder,
            index,
            self.__convention,
            self.__region,
            self.__header,
            new_cache,
            self.__sort_by,
            self.__hidden,
            self.__derived,
        )

    def with_mask(self, mask: NDArray[np.bool_]):
        new_index = self.__index.mask(mask)
        return self.with_index(new_index)

    def make_schema(self, handler: "DatasetHandler"):
        builder_names = set(self.__builder.columns)
        header = self.__header.with_region(self.__region)
        schema = handler.prep_write(self.__index, builder_names - self.__hidden, header)
        derived_names = set(self.__derived.keys()) - self.__hidden
        derived_data = (
            self.select(derived_names)
            .with_units("unitless", None, None)
            .get_data(handler)
        )

        column_units = handler.get_raw_units(builder_names)
        for dn, derived in self.__derived.items():
            column_units[dn] = derived.get_units(column_units)

        for colname in derived_names:
            attrs = {"unit": str(column_units[colname])}
            coldata = derived_data[colname].value
            colschema = ios.ColumnSchema(
                colname, ChunkedIndex.from_size(len(coldata)), coldata, attrs
            )
            schema.add_child(colschema, colname)

        for colname, coldata in self.__im_handler.columns():
            if colname in self.__hidden:
                continue
            attrs = {}
            if isinstance(coldata, units.Quantity):
                attrs["unit"] = str(coldata.unit)
                coldata = coldata.value

            colschema = ios.ColumnSchema(
                colname, ChunkedIndex.from_size(len(coldata)), coldata, attrs
            )
            schema.add_child(colschema, colname)

        return schema

    def with_new_columns(
        self, **new_columns: DerivedColumn | np.ndarray | units.Quantity
    ):
        """
        Add a set of derived columns to the dataset. A derived column is a column that
        has been created based on the values in another column.
        """
        column_names = (
            set(self.builder.columns)
            | set(self.__derived.keys())
            | set(self.__im_handler.keys())
        )
        new_im_handler = self.__im_handler
        derived_update = {}
        for name, new_col in new_columns.items():
            if name in column_names:
                raise ValueError(f"Dataset already has column named {name}")

            if isinstance(new_col, np.ndarray):
                if len(new_col) != len(self.__index):
                    raise ValueError(
                        f"New column {name} has length {len(new_col)} but this dataset "
                        "has length {len(self.__index)}"
                    )
                new_im_handler = new_im_handler.with_new_column(name, new_col)

            elif not new_col.check_parent_existance(column_names):
                raise ValueError(
                    f"Derived column {name} is derived from columns "
                    "that are not in the dataset!"
                )
            else:
                derived_update[name] = new_col
            column_names.add(name)

        new_derived = self.__derived | derived_update
        return DatasetState(
            self.__base_unit_transformations,
            self.__builder,
            self.__index,
            self.__convention,
            self.__region,
            self.__header,
            new_im_handler,
            self.__sort_by,
            self.__hidden,
            new_derived,
        )

    def __build_derived_columns(self, data: table.Table) -> table.Table:
        """
        Build any derived columns that are present in this dataset
        """
        for colname, column in self.__derived.items():
            new_column = column.evaluate(data)
            data[colname] = new_column
        return data

    def __get_im_columns(self, data: table.Table) -> table.Table:
        for colname, column in self.__im_handler.columns():
            data[colname] = column
        return data

    def with_region(self, region: Region):
        """
        Return the same dataset but with a different region
        """
        return DatasetState(
            self.__base_unit_transformations,
            self.__builder,
            self.__index,
            self.__convention,
            region,
            self.__header,
            self.__im_handler,
            self.__sort_by,
            self.__hidden,
            self.__derived,
        )

    def select(self, columns: str | Iterable[str]):
        """
        Select a subset of columns from the dataset. It is possible for a user to select
        a derived column in the dataset, but not the columns it is derived from.
        This class tracks any columns which are required to materialize the dataset but
        are not in the final selection in self.__hidden. When the dataset is
        materialized, the columns in self.__hidden are removed before the data is
        returned to the user.

        """
        if isinstance(columns, str):
            columns = [columns]

        columns = set(columns)
        new_hidden = set()
        if self.__sort_by is not None:
            if self.__sort_by[0] not in columns:
                new_hidden.add(self.__sort_by[0])
            columns.add(self.__sort_by[0])

        known_builders = set(self.__builder.columns)
        known_derived = set(self.__derived.keys())
        known_im = set(self.__im_handler.keys())
        unknown_columns = columns - known_builders - known_derived - known_im
        if unknown_columns:
            raise ValueError(
                "Tried to select columns that aren't in this dataset! Missing columns "
                + ", ".join(unknown_columns)
            )

        required_derived = known_derived.intersection(columns)
        required_builders = known_builders.intersection(columns)
        required_im = known_im.intersection(columns)

        additional_derived = required_derived

        while additional_derived:
            # Follow any chains of derived columns until we reach columns that are
            # actually in the raw data.
            required_derived |= additional_derived
            additional_columns: set[str] = reduce(
                lambda s, derived: s.union(self.__derived[derived].requires()),
                additional_derived,
                set(),
            )
            required_builders |= additional_columns.intersection(known_builders)
            required_im |= additional_columns.intersection(known_im)
            additional_derived = additional_columns.intersection(known_derived)

        all_required = required_derived | required_builders | required_im

        # Derived columns have to be instantiated in the order they are created in order
        # to ensure chains of derived columns work correctly
        new_derived = {k: v for k, v in self.__derived.items() if k in required_derived}
        # Builders can be performed in any order
        new_builder = self.__builder.with_columns(required_builders)
        new_im_handler = self.__im_handler.with_columns(required_im)

        new_hidden.update(all_required - columns)
        if self.__sort_by is not None and self.__sort_by[0] not in columns:
            new_hidden.add(self.__sort_by[0])

        return DatasetState(
            self.__base_unit_transformations,
            new_builder,
            self.__index,
            self.__convention,
            self.__region,
            self.__header,
            new_im_handler,
            self.__sort_by,
            new_hidden,
            new_derived,
        )

    def sort_by(self, column_name: str, handler: "DatasetHandler", invert: bool):
        return DatasetState(
            self.__base_unit_transformations,
            self.__builder,
            self.__index,
            self.__convention,
            self.__region,
            self.__header,
            self.__im_handler,
            (column_name, invert),
            self.__hidden,
            self.__derived,
        )

    def take(self, n: int, at: str, handler):
        """
        Take rows from the dataset.
        """

        if self.__sort_by is not None:
            column = self.select(self.__sort_by[0]).get_data(handler, ignore_sort=True)[
                self.__sort_by[0]
            ]
            sorted = np.argsort(column)
            if self.__sort_by[1]:
                sorted = sorted[::-1]

            index: DataIndex = SimpleIndex(sorted)
        else:
            index = self.__index

        new_index = index.take(n, at)
        if self.__sort_by is not None:
            new_idxs = self.__index.into_array()[new_index.into_array()]
            new_index = SimpleIndex(np.sort(new_idxs))

        return self.with_index(new_index)

    def take_range(self, start: int, end: int):
        """
        Take a range of rows form the dataset.
        """
        if start < 0 or end < 0:
            raise ValueError("start and end must be positive.")
        if end < start:
            raise ValueError("end must be greater than start.")
        if end > len(self.__index):
            raise ValueError("end must be less than the length of the dataset.")

        if start < 0 or end > len(self.__index):
            raise ValueError("start and end must be within the bounds of the dataset.")

        new_index = self.__index.take_range(start, end)
        return self.with_index(new_index)

    def with_units(
        self, convention: str, cosmology: Cosmology, redshift: float | table.Column
    ):
        """
        Change the unit convention
        """
        new_transformations = u.get_unit_transition_transformations(
            self.__header.file.unit_convention,
            convention,
            self.__base_unit_transformations,
            cosmology,
            redshift,
        )
        convention_ = u.UnitConvention(convention)
        new_builder = get_table_builder(new_transformations, self.__builder.columns)
        return DatasetState(
            self.__base_unit_transformations,
            new_builder,
            self.__index,
            convention_,
            self.__region,
            self.__header,
            self.__im_handler,
            self.__sort_by,
            self.__hidden,
            self.__derived,
        )



from typing import TypeGuard

import h5py
import numpy as np
from numpy.typing import NDArray

from opencosmo.index import simple
from opencosmo.index.protocols import DataIndex


def pack(start: np.ndarray, size: np.ndarray) -> tuple[np.ndarray, np.ndarray]:
    """
    Combine adjacent chunks into a single chunk.
    """

    # Calculate the end of each chunk
    end = start + size

    # Determine where a new chunk should start (i.e., not adjacent to previous)
    # We prepend True for the first chunk to always start a group
    new_group = np.ones(len(start), dtype=bool)
    new_group[1:] = start[1:] != end[:-1]

    # Assign a group ID for each segment
    group_ids = np.cumsum(new_group)

    # Combine chunks by group
    combined_start = np.zeros(group_ids[-1], dtype=start.dtype)
    combined_size = np.zeros_like(combined_start)

    np.add.at(combined_start, group_ids - 1, np.where(new_group, start, 0))
    np.add.at(combined_size, group_ids - 1, size)

    return combined_start, combined_size


class ChunkedIndex:
    def __init__(self, starts: np.ndarray, sizes: np.ndarray) -> None:
        self.__starts = starts
        self.__sizes = sizes

    def range(self) -> tuple[int, int]:
        """
        Get the range of the index.
        """
        return self.__starts[0], self.__starts[-1] + self.__sizes[-1] - 1

    def into_array(self) -> NDArray[np.int_]:
        """
        Convert the ChunkedIndex to a SimpleIndex.
        """
        if len(self) == 0:
            return np.array([], dtype=int)
        idxs = np.concatenate(
            [
                np.arange(start, start + size)
                for start, size in zip(self.__starts, self.__sizes)
            ]
        )
        return np.unique(idxs)

    def into_mask(self) -> np.ndarray:
        ends = self.__starts + self.__sizes
        mask = np.zeros(np.max(ends), dtype=bool)
        for start, end in np.nditer([self.__starts, ends]):
            mask[start:end] = True
        return mask

    def intersection(self, other: DataIndex) -> DataIndex:
        if len(self) == 0 or len(other) == 0:
            return simple.SimpleIndex.empty()
        other_mask = other.into_mask()
        self_mask = self.into_mask()
        length = max(len(other_mask), len(self_mask))
        self_mask.resize(length)
        other_mask.resize(length)
        new_idx = np.where(self_mask & other_mask)[0]
        return simple.SimpleIndex(new_idx)

    def projection(self, other: DataIndex) -> DataIndex:
        self_idx = self.into_array()
        other_idx = other.into_array()
        idxs = np.searchsorted(self_idx, other_idx)
        idxs = idxs[idxs != len(self_idx)]
        idxs = idxs[other_idx == self_idx[idxs]]
        return simple.SimpleIndex(idxs)

    def concatenate(self, *others: DataIndex) -> DataIndex:
        if len(others) == 0:
            return self

        if all_are_chunked(others):
            new_starts = np.concatenate(
                [self.__starts] + [other.__starts for other in others]
            )
            new_sizes = np.concatenate(
                [self.__sizes] + [other.__sizes for other in others]
            )
            return ChunkedIndex(new_starts, new_sizes)

        else:
            indexes = np.concatenate([o.into_array() for o in others])
            indexes = np.concatenate(self.into_array(), indexes)
            return simple.SimpleIndex(indexes)

    @classmethod
    def from_size(cls, size: int) -> "ChunkedIndex":
        """
        Create a ChunkedIndex from a size.
        """
        if size <= 0:
            return ChunkedIndex.empty()
        # Create an array of chunk sizes

        starts = np.array([0])
        sizes = np.array([size])
        return ChunkedIndex(starts, sizes)

    @classmethod
    def single_chunk(cls, start: int, size: int) -> "ChunkedIndex":
        """
        Create a ChunkedIndex with a single chunk.
        """
        if start < 0:
            raise ValueError(f"Start must be non-negative, got {start}")

        if not size:
            return ChunkedIndex.empty()
        starts = np.array([start])
        sizes = np.array([size])
        return ChunkedIndex(starts, sizes)

    @classmethod
    def empty(cls):
        start = np.array([], dtype=int)
        size = np.array([], dtype=int)
        return ChunkedIndex(start, size)

    def set_data(self, data: np.ndarray, value: bool) -> np.ndarray:
        """
        Set the data at the index to the given value.
        """
        if not isinstance(data, np.ndarray):
            raise ValueError("Data must be a numpy array")

        for start, size in zip(self.__starts, self.__sizes):
            data[start : start + size] = value
        return data

    def __len__(self) -> int:
        """
        Get the total size of the index.
        """
        return np.sum(self.__sizes)

    def take(self, n: int, at: str = "random") -> DataIndex:
        if n > len(self):
            raise ValueError(f"Cannot take {n} elements from index of size {len(self)}")

        if at == "random":
            idxs = np.concatenate(
                [
                    np.arange(start, start + size)
                    for start, size in zip(self.__starts, self.__sizes)
                ]
            )
            idxs = np.random.choice(idxs, n, replace=False)
            return simple.SimpleIndex(idxs)

        elif at == "start":
            last_chunk_in_range = np.searchsorted(np.cumsum(self.__sizes), n)
            new_starts = self.__starts[: last_chunk_in_range + 1].copy()
            new_sizes = self.__sizes[: last_chunk_in_range + 1].copy()
            new_sizes[-1] = n - np.sum(new_sizes[:-1])
            return ChunkedIndex(new_starts, new_sizes)

        elif at == "end":
            starting_chunk = np.searchsorted(np.cumsum(self.__sizes), len(self) - n)
            new_sizes = self.__sizes[starting_chunk:].copy()
            new_starts = self.__starts[starting_chunk:].copy()
            new_sizes[0] = n - np.sum(new_sizes[1:])
            new_starts[0] = (
                self.__starts[starting_chunk]
                + self.__sizes[starting_chunk]
                - new_sizes[0]
            )
            return ChunkedIndex(new_starts, new_sizes)
        else:
            raise ValueError(f"Unknown value for 'at': {at}")

    def take_range(self, start: int, end: int) -> DataIndex:
        """
        Take a range of elements from the index.
        """
        if start < 0 or end > len(self):
            raise ValueError(
                f"Range {start}:{end} is out of bounds for index of size {len(self)}"
            )

        if start > end:
            raise ValueError(f"Start {start} must be less than end {end}")

        if start == end:
            return ChunkedIndex.empty()

        # Get the indices of the chunks that are in the range
        cumulative_sizes = np.cumsum(self.__sizes)
        start_idx = np.argmax(cumulative_sizes > start)
        end_idx = np.argmax(cumulative_sizes >= end)

        chunk_starts = self.__starts[start_idx : end_idx + 1].copy()
        chunk_sizes = self.__sizes[start_idx : end_idx + 1].copy()

        chunk_starts[0] = chunk_starts[0] + (
            start - cumulative_sizes[start_idx] + cumulative_sizes[0]
        )
        chunk_sizes[-1] = chunk_sizes[-1] - (np.sum(chunk_sizes) - end + start)

        return ChunkedIndex(chunk_starts, chunk_sizes)

    def mask(self, mask: np.ndarray) -> DataIndex:
        """
        Mask the index with a boolean mask.
        """
        if mask.shape != (len(self),):
            raise ValueError(
                f"Mask shape {mask.shape} does not match index size {len(self)}"
            )

        if mask.dtype != bool:
            raise ValueError(f"Mask dtype {mask.dtype} is not boolean")

        if not mask.any():
            return simple.SimpleIndex.empty()

        if mask.all():
            return self

        # Get the indices of the chunks that are masked
        idxs = np.concatenate(
            [
                np.arange(start, start + size)
                for start, size in zip(self.__starts, self.__sizes)
            ]
        )
        masked_idxs = idxs[mask]

        return simple.SimpleIndex(masked_idxs)

    def get_data(self, data: h5py.Dataset | np.ndarray) -> np.ndarray:
        """
        Get the data from the dataset using the index. We want to perform as few reads
        as possible. However, the chunks may not be continuous. This method sorts the
        chunks so it can read the data in the largest possible chunks, it then
        reshuffles the data to match the original order.

        For large numbers of chunks, this is much much faster than reading each chunk
        in the order they are stored in the index. I know because I tried. It sucked.
        """
        if not isinstance(data, (h5py.Dataset, np.ndarray)):
            raise ValueError("Data must be a h5py.Dataset")

        if len(self) == 0:
            return np.array([], dtype=data.dtype)
        if len(self.__starts) == 1:
            return data[self.__starts[0] : self.__starts[0] + self.__sizes[0]]

        sorted_start_index = np.argsort(self.__starts)
        new_starts = self.__starts[sorted_start_index]
        new_sizes = self.__sizes[sorted_start_index]

        packed_starts, packed_sizes = pack(new_starts, new_sizes)

        shape = (len(self),) + data.shape[1:]
        temp = np.zeros(shape, dtype=data.dtype)
        running_index = 0
        for i, (start, size) in enumerate(zip(packed_starts, packed_sizes)):
            temp[running_index : running_index + size] = data[start : start + size]
            running_index += size
        output = np.zeros(len(self), dtype=data.dtype)
        cumulative_sorted_sizes = np.insert(np.cumsum(new_sizes), 0, 0)
        cumulative_original_sizes = np.insert(np.cumsum(self.__sizes), 0, 0)

        # reshuffle the output to match the original order
        for i, sorted_index in enumerate(sorted_start_index):
            start = cumulative_original_sizes[sorted_index]
            end = cumulative_original_sizes[sorted_index + 1]
            data = temp[cumulative_sorted_sizes[i] : cumulative_sorted_sizes[i + 1]]
            output[start:end] = data

        return output

    def n_in_range(
        self, start: NDArray[np.int_], size: NDArray[np.int_]
    ) -> NDArray[np.int_]:
        """
        Return the number of elements in this index that fall within
        a specified data range. Used to mask spatial index.


        As with numpy, this is the half-open range [start, end)
        """
        if len(start) != len(size):
            raise ValueError("Start and size arrays must have the same length")
        if np.any(size < 0):
            raise ValueError("Sizes must greater than or equal to zero")
        if len(self) == 0:
            return np.zeros_like(start)

        required_memory = len(start) * len(self.__starts) * 8
        MEMORY_LIMIT = 2_000_000_000
        if required_memory > MEMORY_LIMIT:
            chunk_size = MEMORY_LIMIT // (8 * len(self.__starts))
            n_chunks = np.ceil(len(start) / chunk_size)
        else:
            n_chunks = 1

        ends = start + size
        self_ends = self.__starts + self.__sizes

        rs = 0
        output = np.zeros_like(start)
        for arr in np.array_split(start, n_chunks):
            in_range_starts = np.clip(
                self.__starts[:, np.newaxis],
                a_min=arr,
                a_max=ends[rs : rs + len(arr)],
            )
            in_range_ends = np.clip(
                self_ends[:, np.newaxis],
                a_min=arr,
                a_max=ends[rs : rs + len(arr)],
            )
            output[rs : rs + len(arr)] = np.sum(in_range_ends - in_range_starts, axis=0)
            rs += len(arr)

        return output
        # example: row[0] in in_range_starts tells me what happens if I clip the 0th
        # element in self.__starts to each of the values in the start and end array
        # that have been provided in the call.

    def __getitem__(self, item: int) -> DataIndex:
        """
        Get an item from the index.
        """
        if item < 0 or item >= len(self):
            raise IndexError(
                f"Index {item} out of bounds for index of size {len(self)}"
            )
        sums = np.cumsum(self.__sizes)
        index = np.searchsorted(sums, item)
        start = self.__starts[index]
        offset = item - sums[index - 1] if index > 0 else item
        offset = int(offset)
        return simple.SimpleIndex(np.array([start + offset]))


def all_are_chunked(
    others: tuple[DataIndex, ...],
) -> TypeGuard[tuple[ChunkedIndex, ...]]:
    """
    Check if all elements in the tuple are instances of ChunkedIndex.
    """
    return all(isinstance(other, ChunkedIndex) for other in others) or not others



from .generator import generate_transformations
from .protocols import (
    AllColumnTransformation,
    ColumnTransformation,
    TableTransformation,
    Transformation,
    TransformationDict,
    TransformationGenerator,
    TransformationType,
)

__all__ = [
    "AllColumnTransformation",
    "ColumnTransformation",
    "TableTransformation",
    "Transformation",
    "generate_transformations",
    "TransformationGenerator",
    "TransformationDict",
    "TransformationType",
]



from .handler import LinkedDatasetHandler
from .io import open_linked_files
from .protocols import LinkHandler
from .structure import StructureCollection

__all__ = [
    "StructureCollection",
    "LinkHandler",
    "LinkedDatasetHandler",
    "OomLinkHandler",
    "open_linked_files",
]



from typing import Protocol, Self

import h5py
import numpy as np
from numpy.typing import NDArray


class DataIndex(Protocol):
    """
    A DataIndex is responsible for tracking which elements of some hdf5 dataset
    should be read. Note these are NOT a spatial index, though a spatial index
    does return a DataIndex when it is queried.
    """

    @classmethod
    def from_size(cls, size: int) -> Self: ...
    def get_data(self, data: h5py.Dataset | np.ndarray) -> np.ndarray: ...
    def n_in_range(
        self, start: NDArray[np.int_], end: NDArray[np.int_]
    ) -> NDArray[np.int_]: ...
    def take(self, n: int, at: str = "random") -> "DataIndex": ...
    def take_range(self, start: int, end: int) -> "DataIndex": ...
    def intersection(self, other: "DataIndex") -> "DataIndex": ...
    def projection(self, other: "DataIndex") -> "DataIndex": ...
    def into_mask(self) -> NDArray[np.bool_]: ...
    def into_array(self) -> NDArray[np.int_]: ...
    def mask(self, mask: np.ndarray) -> "DataIndex": ...
    def range(self) -> tuple[int, int]: ...
    def concatenate(self, *others: "DataIndex") -> "DataIndex": ...
    def __len__(self) -> int: ...
    def __getitem__(self, item: int) -> "DataIndex": ...



from typing import TYPE_CHECKING

import numpy as np

from opencosmo.index import SimpleIndex

if TYPE_CHECKING:
    from opencosmo.dataset.state import DatasetState


def make_sorted_index(state: "DatasetState", column: np.ndarray, invert: bool):
    if not invert:
        order = np.argsort(-column)
    else:
        order = np.argsort(column)

    existing_index = state.index.into_array()
    new_index = SimpleIndex(existing_index[order])
    return new_index



from __future__ import annotations

from collections.abc import Iterable, Sequence

from astropy.table import Column, QTable  # type: ignore
from numpy.typing import NDArray

import opencosmo.transformations as t


def get_table_builder(
    transformations: t.TransformationDict, column_names: Iterable[str]
) -> "TableBuilder":
    """
    This function creates a dictionary of ColumnBuilders from a dictionary of
    transformations. The keys of the dictionary are the column names and the
    values are the ColumnBuilders.
    """
    column_transformations = transformations.get(t.TransformationType.COLUMN, [])
    all_column_transformations = transformations.get(
        t.TransformationType.ALL_COLUMNS, []
    )
    table_transformations = transformations.get(t.TransformationType.TABLE, [])
    if not all(
        isinstance(transformation, t.AllColumnTransformation)
        for transformation in all_column_transformations
    ):
        raise ValueError("Expected AllColumnTransformation.")
    column_builders: dict[str, list[t.Transformation]] = {
        name: [] for name in column_names
    }
    for transformation in column_transformations:
        if not hasattr(transformation, "column_name"):
            raise ValueError(
                f"Expected ColumnTransformation, got {type(transformation)}."
            )
        if transformation.column_name not in column_builders:
            continue
        column_builders[transformation.column_name].append(transformation)

    for column_name in column_names:
        column_builders[column_name].extend(all_column_transformations)
    columns = {
        name: ColumnBuilder(name, builders)
        for name, builders in column_builders.items()
    }
    return TableBuilder(columns, table_transformations)


def apply_table_transformations(
    table: QTable, transformations: Sequence[t.TableTransformation]
):
    """
    Apply transformations to the table as a whole. These transformations
    are applied after individual column transformations.
    """
    output_table = table
    for tr in transformations:
        if (new_table := tr(output_table)) is not None:
            output_table = new_table
    return output_table


class TableBuilder:
    def __init__(
        self,
        columns: dict[str, "ColumnBuilder"],
        transformations: Sequence[t.Transformation],
    ):
        self.column_builders = columns
        self.transformations = transformations

    @property
    def columns(self):
        return self.column_builders.keys()

    def with_columns(self, columns: Iterable[str]):
        new_columns = {c: self.column_builders[c] for c in columns}
        return TableBuilder(new_columns, self.transformations)

    def build(self, columns: dict[str, NDArray]):
        output_columns = {}
        for name, data in columns.items():
            builder = self.column_builders.get(name)
            if builder is None:
                builder = ColumnBuilder.default(name)
            column = builder.build(data)
            output_columns[name] = column
        table = QTable(output_columns)

        return apply_table_transformations(table, self.transformations)


class ColumnBuilder:
    """
    OpenCosmo operates on columns of data, only producing an actual full Astropy table
    when data is actually requested. Things like filtering, selecting, and changing
    units are repesented as transformations on the given column.

    The handler is responsible for actually getting the data from the source and
    feeding it to the ColumBuilder.
    """

    def __init__(
        self,
        name: str,
        transformations: Sequence[t.Transformation],
    ):
        self.column_name = name
        self.transformations = transformations

    @classmethod
    def default(cls, name):
        return ColumnBuilder(name, {})

    def build(self, data: NDArray):
        """
        The column should always come to the builder without
        units.
        """
        new_column = Column(data)
        for transformation in self.transformations:
            transformed_column = transformation(new_column)
            if transformed_column is not None:
                new_column = transformed_column
        return new_column



from typing import Optional, Type

from astropy.cosmology import Cosmology
from astropy.units.typing import UnitLike
from pydantic import BaseModel

import opencosmo.transformations as t
from opencosmo.transformations.protocols import ColumnTransformation, TransformationType
from opencosmo.transformations.units import (
    apply_unit,
    get_unit_transition_transformations,
)

ModelUnitAnnotation = tuple[str, dict[str, UnitLike]]

__KNOWN_UNITFUL_MODELS__: dict[Type[BaseModel], ModelUnitAnnotation] = {}


# Constraint: Unit covention for all fields in a given model must be the same


def register_units(
    model: Type[BaseModel],
    field_name: str,
    unit: UnitLike,
    convention: str = "scalefree",
):
    model_spec = __KNOWN_UNITFUL_MODELS__.get(model)
    registered_fields: dict[str, UnitLike]
    if model_spec is not None and model_spec[0] != convention:
        raise ValueError(
            "All unitful fields in a parameter model must use the same unit convention"
        )
    elif model_spec is None:
        registered_fields = {}
    else:
        registered_fields = model_spec[1]
    if field_name in registered_fields:
        raise ValueError(f"Field {field_name} was already registered with units!")

    registered_fields[field_name] = unit
    __KNOWN_UNITFUL_MODELS__[model] = (convention, registered_fields)


def __get_unit_transformations(
    model: BaseModel, cosmology, convention: str = "scalefree"
) -> t.TransformationDict:
    transformations: t.TransformationDict = {}
    if (us := __KNOWN_UNITFUL_MODELS__.get(type(model))) is None:
        return {}

    base_convention, known_units = us

    column_transformations: list[ColumnTransformation] = []
    for name, unit in known_units.items():
        column_transformations.append(apply_unit(name, unit))

    transformations[TransformationType.COLUMN] = column_transformations

    transformations = get_unit_transition_transformations(
        base_convention, convention, transformations, cosmology
    )
    return transformations


def apply_units(
    model: BaseModel, cosmology: Optional[Cosmology], convention: str = "scalefree"
):
    transformations = __get_unit_transformations(model, cosmology, convention)
    parameters = model.model_dump()
    column_transformations = transformations.get(TransformationType.COLUMN, [])
    all_column_transformations = transformations.get(TransformationType.ALL_COLUMNS, [])
    for trans in column_transformations:
        assert hasattr(trans, "column_name")
        value = trans(parameters[trans.column_name])

        for trans_ in all_column_transformations:
            value = trans_(value)
        parameters[trans.column_name] = value
    return parameters



from enum import Enum
from functools import partial
from typing import TYPE_CHECKING, Optional

import astropy.cosmology.units as cu  # type: ignore
import astropy.units as u  # type: ignore
import h5py
from astropy.constants import m_p  # type: ignore
from astropy.cosmology import Cosmology
from astropy.table import Column, Table  # type: ignore
from numpy.typing import ArrayLike

from opencosmo import transformations as t
from opencosmo.dataset.builders import get_table_builder

if TYPE_CHECKING:
    from opencosmo.header import OpenCosmoHeader

_ = u.add_enabled_units(cu)

KNOWN_REDSHIFT_COLUMNS = {
    "redshift",
    "redshift_true",
}

KNOWN_SCALE_COLUMS = {"fof_halo_center_a"}


def get_a(table: Table):
    columns = set(table.columns)
    redshift_columns = columns.intersection(KNOWN_REDSHIFT_COLUMNS)
    if len(redshift_columns) == 1:
        redshifts = table[redshift_columns.pop()]
        return 1 / (redshifts + 1)

    scale_columns = columns.intersection(KNOWN_SCALE_COLUMS)

    if len(scale_columns) == 1:
        return table[scale_columns.pop()]
    raise ValueError("Unable to find redshift column to convert coordinates")


UNIT_MAP = {
    "comoving Mpc/h": u.Mpc / cu.littleh,
    "comoving (Mpc/h)^2": (u.Mpc / cu.littleh) ** 2,
    "comoving km/s": u.km / u.s,
    "comoving (km/s)^2": (u.km / u.s) ** 2,
    "Msun/h": u.Msun / cu.littleh,
    "Msun/yr": u.Msun / u.yr,
    "K": u.K,
    "comoving (Msun/h * (km/s) * Mpc/h)": (u.Msun / cu.littleh)
    * (u.km / u.s)
    * (u.Mpc / cu.littleh),
    "log10(erg/s)": u.DexUnit("erg/s"),
    "h^2 keV / (comoving cm)^3": (cu.littleh**2) * u.keV / (u.cm**3),
    "keV * cm^2": u.keV * u.cm**2,
    "cm^-3": u.cm**-3,
    "Gyr": u.Gyr,
    "Msun/h / (comoving Mpc/h)^3": (u.Msun / cu.littleh) / (u.Mpc / cu.littleh) ** 3,
    "Msun/h * km/s": (u.Msun / cu.littleh) * (u.km / u.s),
    "H0^-1": (u.s * (1 * u.Mpc).to(u.km).value).to(u.year) / (100 * cu.littleh),
    "m_hydrogen": m_p,
    "Msun * (km/s)^2": (u.Msun) * (u.km / u.s) ** 2,
}


class UnitError(Exception):
    pass


class UnitConvention(Enum):
    COMOVING = "comoving"
    PHYSICAL = "physical"
    SCALEFREE = "scalefree"
    UNITLESS = "unitless"


def get_unit_transformation_generators() -> list[t.TransformationGenerator]:
    """
    Get the unit transformation generato.

    We use generators for units because it is most appropriate to think
    of units as fundamental to the data, even when they don't actually
    appear in the hdf5 file.

    Even if the user requests unitless data, we still need to have
    access to these so that we can apply units if they
    call Dataset.with_convention later.
    """
    return [
        generate_attribute_unit_transformations,
    ]


def get_unit_transition_transformations(
    from_convention: str | UnitConvention,
    to_convention: str | UnitConvention,
    unit_applicators: t.TransformationDict,
    cosmology: Cosmology,
    redshift: float = 0,
) -> t.TransformationDict:
    """
    Given a dataset, the user can request a transformation to a different unit
    convention. The returns a new set of transformations that will take the
    dataset to the requested unit convention.
    """
    from_ = UnitConvention(from_convention)
    to_ = UnitConvention(to_convention)
    if to_ == UnitConvention.UNITLESS:
        return {}
    match from_:
        case UnitConvention.SCALEFREE:
            update_transformations = get_scalefree_transition_transformations(
                to_, unit_applicators, cosmology, redshift
            )
        case UnitConvention.COMOVING:
            update_transformations = get_comoving_transition_transformations(
                to_, unit_applicators, cosmology, redshift
            )
        case UnitConvention.PHYSICAL:
            update_transformations = get_physical_transition_transformations(
                to_, unit_applicators, cosmology, redshift
            )
        case UnitConvention.UNITLESS:
            raise UnitError("Cannot convert units for datasets that have none!")

    for ttype in unit_applicators:
        existing = update_transformations.get(ttype, [])
        update_transformations[ttype] = unit_applicators[ttype] + existing
    return update_transformations


def get_comoving_transition_transformations(
    to: UnitConvention,
    unit_applicators: t.TransformationDict,
    cosmology: Cosmology,
    redshift: float,
):
    comoving_to_phys: t.TableTransformation = partial(
        comoving_to_physical, cosmology=cosmology, redshift=redshift
    )
    match to:
        case UnitConvention.PHYSICAL:
            return {t.TransformationType.TABLE: [comoving_to_phys]}
        case UnitConvention.COMOVING:
            return {}
        case UnitConvention.SCALEFREE:
            raise UnitError(
                "Data stored in comoving units cannot be converted into scalefree units"
            )


def get_physical_transition_transformations(
    to: UnitConvention,
    unit_applicators: t.TransformationDict,
    cosmology: Cosmology,
    redshift: float,
):
    phys_to_comoving: t.TableTransformation = partial(
        physical_to_comoving, cosmology=cosmology, redshift=redshift
    )
    match to:
        case UnitConvention.PHYSICAL:
            return {}
        case UnitConvention.COMOVING:
            return {t.TransformationType.TABLE: [phys_to_comoving]}
        case UnitConvention.SCALEFREE:
            raise UnitError(
                "Data stored in physical units cannot be converted into scalefree units"
            )


def get_scalefree_transition_transformations(
    to: UnitConvention,
    unit_applicators: t.TransformationDict,
    cosmology: Cosmology,
    redshift: float,
):
    remove_h: t.TableTransformation = partial(remove_littleh, cosmology=cosmology)
    comoving_to_phys: t.TableTransformation = partial(
        comoving_to_physical, cosmology=cosmology, redshift=redshift
    )
    match to:
        case UnitConvention.COMOVING:
            return {t.TransformationType.ALL_COLUMNS: [remove_h]}
        case UnitConvention.PHYSICAL:
            return {
                t.TransformationType.ALL_COLUMNS: [remove_h],
                t.TransformationType.TABLE: [comoving_to_phys],
            }
        case UnitConvention.SCALEFREE:
            return {}


def get_default_unit_transformations(
    file: h5py.File | h5py.Group, header: "OpenCosmoHeader"
):
    data_convention = UnitConvention(header.file.unit_convention)
    unit_applicators = get_base_unit_transformations(file["data"], header)
    match data_convention:
        case UnitConvention.SCALEFREE:
            base_transformations = get_unit_transition_transformations(
                data_convention,
                "comoving",
                unit_applicators,
                header.cosmology,
                header.file.redshift,
            )
        case UnitConvention.UNITLESS:
            base_transformations = {}
        case UnitConvention.COMOVING:
            base_transformations = unit_applicators
        case UnitConvention.PHYSICAL:
            raise NotImplementedError()

    column_names = list(str(col) for col in file["data"].keys())
    builder = get_table_builder(base_transformations, column_names)

    return builder, unit_applicators


def get_base_unit_transformations(
    input: h5py.Dataset,
    header: "OpenCosmoHeader",
) -> t.TransformationDict:
    """
    Get the base unit transformations for a given dataset. These transformations
    produce the units that the data are actually stored in. Datasets alwyas
    hold onto a copy of these transformations even if the user later requests
    a different unit convention.


    These always apply after the initial transformations generated above.
    """
    generators = get_unit_transformation_generators()
    base_transformations = t.generate_transformations(input, generators, {})
    return base_transformations


def remove_littleh(
    column: Column | u.Quantity, cosmology: Cosmology
) -> Optional[Table]:
    """
    Remove little h from the units of the input table. For comoving
    coordinates, this is the second step after parsing the units themselves.
    """
    if (unit := column.unit) is not None:
        # Handle dex units
        try:
            if isinstance(unit, u.DexUnit):
                u_base = unit.physical_unit
                constructor = u.DexUnit
            else:
                u_base = unit

                def constructor(x):
                    return x
        except AttributeError:
            return None

        try:
            index = u_base.bases.index(cu.littleh)
        except ValueError:
            return None
        power = u_base.powers[index]
        new_unit = constructor(u_base / cu.littleh**power)
        column = column.to(new_unit, cu.with_H0(cosmology.H0))
    return column


def comoving_to_physical(
    table: Table, cosmology: Cosmology, redshift: float
) -> Optional[Table]:
    """
    Convert comoving coordinates to physical coordinates. This is the
    second step after parsing the units themselves.
    """
    try:
        a = get_a(table)
    except ValueError:
        a = cosmology.scale_factor(redshift)

    for colname in table.columns:
        if (unit := table[colname].unit) is not None:
            # Check if the units have distances in them
            decomposed = unit.decompose()
            try:
                index = decomposed.bases.index(u.m)
            except (ValueError, AttributeError):
                continue
            power = decomposed.powers[index]
            # multiply by the scale factor to the same power as the distance
            table[colname] = table[colname] * a**power

    return table


def physical_to_comoving(
    table: Table, cosmology: Cosmology, redshift: float
) -> Optional[Table]:
    """
    Convert comoving coordinates to physical coordinates. This is the
    second step after parsing the units themselves.
    """

    try:
        a = get_a(table)
    except ValueError:
        a = cosmology.scale_factor(redshift)

    for colname in table.columns:
        if (unit := table[colname].unit) is not None:
            # Check if the units have distances in them
            decomposed = unit.decompose()
            try:
                index = decomposed.bases.index(u.m)
            except (ValueError, AttributeError):
                continue
            power = decomposed.powers[index]
            # divide by the scale factor to the same power as the distance
            table[colname] = table[colname] / a**power

    return table


def get_raw_units(column: h5py.Dataset):
    if "unit" in column.attrs:
        if (us := column.attrs["unit"]) == "None" or us == "":
            return None
        if (unit := UNIT_MAP.get(us)) is not None:
            return unit
        try:
            return u.Unit(us)
        except ValueError:
            return None

    return None


def generate_attribute_unit_transformations(
    column: h5py.Dataset,
) -> t.TransformationDict:
    """
    Check the attributes of an hdf5 dataset to see if information about units is stored
    there.

    The raw HACC data does not store units in this way, relying instead on standard
    naming conventions. However if a user creates a new column, we want to be able to
    store it in our standard format without losing unit information and we cannot rely
    on them following our naming conventions.
    """
    unit = get_raw_units(column)
    if unit is not None:
        apply_func: t.Transformation = apply_unit(
            column_name=column.name.split("/")[-1], unit=unit
        )
        return {t.TransformationType.COLUMN: [apply_func]}
    return {}


class apply_unit:
    """
    Apply a unit to an input column. Ensuring that the correct column is
    passed will be the responsibility of the caller.

    Has to be a class so it can implement the ColumnTransformation protocol.
    """

    def __init__(self, column_name: str, unit: u.Unit):
        self.__name = column_name
        self.unit = unit

    def __call__(self, input: ArrayLike) -> ArrayLike:
        return input * self.unit

    @property
    def column_name(self) -> str:
        return self.__name
